from remotemanager.connection.computers.summer import Summer
from remotemanager.connection.computers.vega import Vega
from remotemanager.connection.computers.base import BaseComputer, \
    optional, required

__all__ = ['BaseComputer',
           'optional',
           'required',
           'Summer',
           'Vega']
